#!/usr/bin/python3

import os
from typing import List, Union
from deezspot.libutils.utils import sanitize_name
from deezspot.libutils.logging_utils import logger
from deezspot.models.download import Track


def create_m3u_file(output_dir: str, playlist_name: str) -> str:
    """
    Creates an m3u playlist file with the proper header.
    Returns full path to the m3u file.
    """
    playlist_m3u_dir = os.path.join(output_dir, "playlists")
    os.makedirs(playlist_m3u_dir, exist_ok=True)
    playlist_name_sanitized = sanitize_name(playlist_name)
    m3u_path = os.path.join(playlist_m3u_dir, f"{playlist_name_sanitized}.m3u")
    # Always ensure header exists (idempotent)
    if not os.path.exists(m3u_path):
        with open(m3u_path, "w", encoding="utf-8") as m3u_file:
            m3u_file.write("#EXTM3U\n")
        logger.debug(f"Created m3u playlist file: {m3u_path}")
    return m3u_path


def ensure_m3u_header(m3u_path: str) -> None:
    """Ensure an existing m3u has the header; create if missing."""
    if not os.path.exists(m3u_path):
        os.makedirs(os.path.dirname(m3u_path), exist_ok=True)
        with open(m3u_path, "w", encoding="utf-8") as m3u_file:
            m3u_file.write("#EXTM3U\n")


# Prefer the actual file that exists on disk; if the stored path doesn't exist,
# attempt to find the same basename with a different extension (e.g., due to conversion).
_AUDIO_EXTS_TRY = [
    ".flac", ".mp3", ".m4a", ".aac", ".alac", ".ogg", ".opus", ".wav", ".aiff"
]


def _resolve_existing_song_path(song_path: str) -> Union[str, None]:
    if not song_path:
        return None
    if os.path.exists(song_path):
        return song_path
    base, _ = os.path.splitext(song_path)
    for ext in _AUDIO_EXTS_TRY:
        candidate = base + ext
        if os.path.exists(candidate):
            return candidate
    return None


def _get_track_duration_seconds(track: Track) -> int:
    try:
        if hasattr(track, 'tags') and track.tags:
            if 'duration' in track.tags:
                return int(float(track.tags['duration']))
            elif 'length' in track.tags:
                return int(float(track.tags['length']))
        if hasattr(track, 'song_metadata') and hasattr(track.song_metadata, 'duration_ms'):
            return int(track.song_metadata.duration_ms / 1000)
        return 0
    except (ValueError, AttributeError, TypeError):
        return 0


def _get_track_info(track: Track) -> tuple:
    try:
        if hasattr(track, 'tags') and track.tags:
            artist = track.tags.get('artist', 'Unknown Artist')
            title = track.tags.get('music', track.tags.get('title', 'Unknown Title'))
            return artist, title
        elif hasattr(track, 'song_metadata'):
            sep = ", "
            if hasattr(track, 'tags') and track.tags:
                sep = track.tags.get('artist_separator', sep)
            if hasattr(track.song_metadata, 'artists') and track.song_metadata.artists:
                artist = sep.join([a.name for a in track.song_metadata.artists])
            else:
                artist = 'Unknown Artist'
            title = getattr(track.song_metadata, 'title', 'Unknown Title')
            return artist, title
        else:
            return 'Unknown Artist', 'Unknown Title'
    except (AttributeError, TypeError):
        return 'Unknown Artist', 'Unknown Title'


def _read_m3u_entries(m3u_path: str) -> List[tuple]:
    """Parse existing m3u into a list of (extinf_line, path_line) entries after header."""
    entries: List[tuple] = []
    if not os.path.exists(m3u_path):
        return entries
    try:
        with open(m3u_path, "r", encoding="utf-8") as f:
            lines = [line.rstrip('\n') for line in f]
    except OSError:
        return entries
    # Skip header if present
    idx = 0
    if idx < len(lines) and lines[idx].strip() == "#EXTM3U":
        idx += 1
    while idx < len(lines):
        line = lines[idx]
        if not line:
            idx += 1
            continue
        if line.startswith("#EXTINF:"):
            extinf = line
            path = ""
            if idx + 1 < len(lines):
                path = lines[idx + 1]
            entries.append((extinf, path))
            idx += 2
        else:
            # Path-only entry
            entries.append(("", line))
            idx += 1
    return entries


def _write_m3u_entries(m3u_path: str, entries: List[tuple]) -> None:
    """Write header and provided entries back to file."""
    # Ensure folder exists
    os.makedirs(os.path.dirname(m3u_path), exist_ok=True)
    with open(m3u_path, "w", encoding="utf-8") as m3u_file:
        m3u_file.write("#EXTM3U\n")
        for extinf, path in entries:
            # Skip empty placeholders
            if not path and not extinf:
                continue
            if extinf:
                m3u_file.write(f"{extinf}\n")
            if path:
                m3u_file.write(f"{path}\n")


def append_track_to_m3u(m3u_path: str, track: Union[str, Track]) -> None:
    """Append a single track to m3u with EXTINF and a resolved path.
    Idempotent behavior: if entry for same path exists, it is updated/moved to the desired position; if an entry exists at the desired position but differs, it is overwritten.
    """
    ensure_m3u_header(m3u_path)
    # Prepare entries and base dir
    playlist_m3u_dir = os.path.dirname(m3u_path)
    entries = _read_m3u_entries(m3u_path)

    # Handle simple string path case: dedupe by path, append if new
    if isinstance(track, str):
        resolved = _resolve_existing_song_path(track)
        if not resolved:
            return
        relative_path = os.path.relpath(resolved, start=playlist_m3u_dir)
        # Remove existing entries with same path
        new_entries = [(e, p) for (e, p) in entries if p != relative_path]
        # If nothing changed and path already existed identically, skip write
        if len(new_entries) == len(entries) and any(p == relative_path for _, p in entries):
            return
        new_entries.append(("", relative_path))
        _write_m3u_entries(m3u_path, new_entries)
        return

    # Validate Track object
    if (not isinstance(track, Track) or 
        not track.success or 
        not hasattr(track, 'song_path')):
        return

    resolved = _resolve_existing_song_path(track.song_path)
    if not resolved:
        return

    relative_path = os.path.relpath(resolved, start=playlist_m3u_dir)
    duration = _get_track_duration_seconds(track)
    artist, title = _get_track_info(track)
    extinf_line = f"#EXTINF:{duration},{artist} - {title}"
    new_entry = (extinf_line, relative_path)

    # Determine target playlist position (1-based). Fallback to 0 (append behavior) if missing/invalid.
    position = 0
    try:
        if hasattr(track, 'tags') and track.tags:
            raw_pos = track.tags.get('playlistnum')
            if raw_pos is not None:
                position = int(raw_pos)
    except (ValueError, TypeError):
        position = 0

    # Remove duplicates by path first (to avoid multiple occurrences upon re-download)
    entries = [(e, p) for (e, p) in entries if p != relative_path]

    if position >= 1 and position - 1 < len(entries):
        # If there is already something at that position, only overwrite if different
        current = entries[position - 1]
        if current != new_entry:
            entries[position - 1] = new_entry
        else:
            # Exact match at the correct position: nothing to do
            return
    else:
        # If position beyond current length or not provided: append unless identical exists (path dedup above already handled)
        # Avoid appending if an identical entry already exists (by both extinf and path)
        if any(e == new_entry for e in entries):
            return
        # If position is within range but entries is shorter, we won't insert placeholders; we simply append to end.
        entries.append(new_entry)

    _write_m3u_entries(m3u_path, entries)


def write_tracks_to_m3u(output_dir: str, playlist_name: str, tracks: List[Track]) -> str:
    """
    Legacy batch method. Creates an m3u and writes provided tracks.
    Prefer progressive usage: create_m3u_file(...) once, then append_track_to_m3u(...) per track.
    """
    playlist_m3u_dir = os.path.join(output_dir, "playlists")
    os.makedirs(playlist_m3u_dir, exist_ok=True)
    m3u_path = os.path.join(playlist_m3u_dir, f"{sanitize_name(playlist_name)}.m3u")
    ensure_m3u_header(m3u_path)
    for track in tracks:
        append_track_to_m3u(m3u_path, track)
    logger.info(f"Created m3u playlist file at: {m3u_path}")
    return m3u_path


def get_m3u_path(output_dir: str, playlist_name: str) -> str:
    playlist_m3u_dir = os.path.join(output_dir, "playlists")
    return os.path.join(playlist_m3u_dir, f"{sanitize_name(playlist_name)}.m3u") 


#!/usr/bin/python3

from dataclasses import dataclass, field
from typing import Optional, Dict, Any, List


def _str(v: Any) -> Optional[str]:
	if v is None:
		return None
	try:
		s = str(v)
		return s
	except Exception:
		return None


def _int(v: Any) -> Optional[int]:
	try:
		if v is None:
			return None
		return int(v)
	except Exception:
		return None


def _bool(v: Any) -> Optional[bool]:
	if isinstance(v, bool):
		return v
	if v in ("true", "True", "1", 1):
		return True
	if v in ("false", "False", "0", 0):
		return False
	return None


def _list_str(v: Any) -> Optional[List[str]]:
	if v is None:
		return []
	if isinstance(v, list):
		out: List[str] = []
		for it in v:
			s = _str(it)
			if s is not None:
				out.append(s)
		return out
	return []


@dataclass
class ExternalUrls:
	spotify: Optional[str] = None

	@staticmethod
	def from_dict(obj: Any) -> "ExternalUrls":
		if not isinstance(obj, dict):
			return ExternalUrls()
		return ExternalUrls(
			spotify=_str(obj.get("spotify"))
		)

	def to_dict(self) -> Dict[str, Any]:
		return {"spotify": self.spotify} if self.spotify else {}


@dataclass
class Image:
	url: str
	width: int = 0
	height: int = 0

	@staticmethod
	def from_dict(obj: Any) -> Optional["Image"]:
		if not isinstance(obj, dict):
			return None
		url = _str(obj.get("url"))
		if not url:
			return None
		w = _int(obj.get("width")) or 0
		h = _int(obj.get("height")) or 0
		return Image(url=url, width=w, height=h)

	def to_dict(self) -> Dict[str, Any]:
		return {"url": self.url, "width": self.width, "height": self.height}


@dataclass
class ArtistRef:
	id: Optional[str] = None
	name: str = ""
	type: str = "artist"
	uri: Optional[str] = None
	external_urls: ExternalUrls = field(default_factory=ExternalUrls)

	@staticmethod
	def from_dict(obj: Any) -> "ArtistRef":
		if not isinstance(obj, dict):
			return ArtistRef()
		return ArtistRef(
			id=_str(obj.get("id")),
			name=_str(obj.get("name")) or "",
			type=_str(obj.get("type")) or "artist",
			uri=_str(obj.get("uri")),
			external_urls=ExternalUrls.from_dict(obj.get("external_urls", {})),
		)

	def to_dict(self) -> Dict[str, Any]:
		out = {
			"id": self.id,
			"name": self.name,
			"type": self.type,
			"uri": self.uri,
			"external_urls": self.external_urls.to_dict()
		}
		return {k: v for k, v in out.items() if v not in (None, {}, [], "")}


@dataclass
class AlbumRef:
	id: Optional[str] = None
	name: Optional[str] = None
	type: str = "album"
	uri: Optional[str] = None
	external_urls: ExternalUrls = field(default_factory=ExternalUrls)
	images: Optional[List[Image]] = None

	@staticmethod
	def from_dict(obj: Any) -> "AlbumRef":
		if not isinstance(obj, dict):
			return AlbumRef()
		imgs: List[Image] = []
		for im in obj.get("images", []) or []:
			im_obj = Image.from_dict(im)
			if im_obj:
				imgs.append(im_obj)
		return AlbumRef(
			id=_str(obj.get("id")),
			name=_str(obj.get("name")),
			type=_str(obj.get("type")) or "album",
			uri=_str(obj.get("uri")),
			external_urls=ExternalUrls.from_dict(obj.get("external_urls", {})),
			images=imgs or None,
		)

	def to_dict(self) -> Dict[str, Any]:
		out = {
			"id": self.id,
			"name": self.name,
			"type": self.type,
			"uri": self.uri,
			"external_urls": self.external_urls.to_dict(),
			"images": [im.to_dict() for im in (self.images or [])]
		}
		return {k: v for k, v in out.items() if v not in (None, {}, [], "")} 


#!/usr/bin/python3

from deezspot.models.download.track import Track
from deezspot.models.download.album import Album
from deezspot.models.download.playlist import Playlist

class Smart:
	def __init__(self) -> None:
		self.track: Track = None
		self.album: Album = None
		self.playlist: Playlist = None
		self.type = None
		self.source = None


#!/usr/bin/python3
import os
import json
import logging
import re
from deezspot.deezloader.dee_api import API
from deezspot.easy_spoty import Spo
from deezspot.deezloader.deegw_api import API_GW
from deezspot.deezloader.deezer_settings import stock_quality
from deezspot.models.download import (
    Track,
    Album,
    Playlist,
    Preferences,
    Smart,
    Episode,
)
from deezspot.deezloader.__download__ import (
    DW_TRACK,
    DW_ALBUM,
    DW_PLAYLIST,
    DW_EPISODE,
    Download_JOB,
)
from deezspot.exceptions import (
    InvalidLink,
    TrackNotFound,
    NoDataApi,
    AlbumNotFound,
    MarketAvailabilityError,
)
from deezspot.libutils.utils import (
    create_zip,
    get_ids,
    link_is_valid,
    what_kind,
    sanitize_name
)
from deezspot.libutils.others_settings import (
    stock_output,
    stock_recursive_quality,
    stock_recursive_download,
    stock_not_interface,
    stock_zip,
    stock_save_cover,
    stock_market
)
from deezspot.libutils.logging_utils import ProgressReporter, logger, report_progress
import requests
from librespot.core import Session

from deezspot.models.callback.callbacks import (
    trackCallbackObject,
    albumCallbackObject,
    playlistCallbackObject,
    errorObject,
    summaryObject,
    failedTrackObject,
    initializingObject,
    doneObject,
)
from deezspot.models.callback.track import trackObject as trackCbObject, artistTrackObject
from deezspot.models.callback.album import albumObject as albumCbObject
from deezspot.models.callback.playlist import playlistObject as playlistCbObject
from deezspot.models.callback.common import IDs
from deezspot.models.callback.user import userObject
from rapidfuzz import fuzz

def _sim(a: str, b: str) -> float:
    a = (a or '').strip().lower()
    b = (b or '').strip().lower()
    if not a or not b:
        return 0.0
    return fuzz.partial_ratio(a, b) / 100

# Clean for searching on Deezer
def _remove_parentheses(string: str) -> str:
    # remove () and [] and {}, as well as anything inside
    return re.sub(r'\{[^)]*\}', '', re.sub(r'\[[^)]*\]', '', re.sub(r'\([^)]*\)', '', string)))

API()

# Create a logger for the deezspot library
logger = logging.getLogger('deezspot')

class DeeLogin:
    def __init__(
        self,
        arl=None,
        email=None,
        password=None,
        spotify_client_id=None,
        spotify_client_secret=None,
        spotify_credentials_path=None,
        progress_callback=None,
        silent=False
    ) -> None:

        # Store Spotify credentials
        self.spotify_client_id = spotify_client_id
        self.spotify_client_secret = spotify_client_secret
        # Optional path to Spotify credentials.json (explicit param > env override > CWD default)
        self.spotify_credentials_path = (
            spotify_credentials_path
            or os.environ.get("SPOTIFY_CREDENTIALS_PATH")
            or os.path.join(os.getcwd(), "credentials.json")
        )
        
        # Initialize Spotify API if credentials are provided
        if spotify_client_id and spotify_client_secret:
            Spo.__init__(client_id=spotify_client_id, client_secret=spotify_client_secret)

        # Initialize Deezer API
        if arl:
            self.__gw_api = API_GW(arl=arl)
        else:
            self.__gw_api = API_GW(
                email=email,
                password=password
            )
            
        # Reference to the Spotify search functionality
        self.__spo = Spo
        
        # Configure progress reporting
        self.progress_reporter = ProgressReporter(callback=progress_callback, silent=silent)
        
        # Set the progress reporter for Download_JOB
        Download_JOB.set_progress_reporter(self.progress_reporter)

    def _ensure_spotify_session(self) -> None:
        """Ensure Spo has an attached librespot Session. Used only by spo->dee flows."""
        try:
            # Check if Spo already has a session (accessing private attr is ok internally)
            has_session = getattr(Spo, f"_{Spo.__name__}__session", None) is not None
            if has_session:
                return
        except Exception:
            pass

        cred_path = self.spotify_credentials_path
        if not os.path.isfile(cred_path):
            raise FileNotFoundError(
                f"Spotify session not initialized. Missing credentials.json at '{cred_path}'. "
                "Set SPOTIFY_CREDENTIALS_PATH or place credentials.json in the working directory."
            )
        builder = Session.Builder()
        builder.conf.stored_credentials_file = cred_path
        session = builder.stored_file().create()
        Spo.set_session(session)

    def download_trackdee(
        self, link_track,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        playlist_context=None,
        artist_separator: str = "; ",
        spotify_metadata: bool = False,
        pad_number_width: int | str = 'auto'
    ) -> Track:

        link_is_valid(link_track)
        ids = get_ids(link_track)
        track_obj = None

        def report_error(e, current_ids, url):
            error_status = errorObject(ids=IDs(deezer=current_ids), error=str(e))
            summary = summaryObject(
                failed_tracks=[failedTrackObject(track=trackCbObject(title=f"Track ID {current_ids}"), reason=str(e))],
                total_failed=1
            )
            error_status.summary = summary
            callback_obj = trackCallbackObject(
                track=trackCbObject(title=f"Track ID {current_ids}", ids=IDs(deezer=current_ids)),
                status_info=error_status
            )
            report_progress(reporter=self.progress_reporter, callback_obj=callback_obj)

        try:
            # Default: Get standardized Deezer track object for tagging
            track_obj = API.get_track(ids)
        except (NoDataApi, MarketAvailabilityError) as e:
            # Try to get fallback track information
            infos = self.__gw_api.get_song_data(ids)
            if "FALLBACK" not in infos:
                report_error(e, ids, link_track)
                raise TrackNotFound(link_track) from e

            fallback_id = infos['FALLBACK']['SNG_ID']
            try:
                # Try again with fallback ID
                track_obj = API.get_track(fallback_id)
                if not track_obj or not track_obj.available:
                    raise MarketAvailabilityError(f"Fallback track {fallback_id} not available.")
                # Update the ID to use the fallback
                ids = fallback_id
            except (NoDataApi, MarketAvailabilityError) as e_fallback:
                report_error(e_fallback, fallback_id, link_track)
                raise TrackNotFound(url=link_track, message=str(e_fallback)) from e_fallback
        
        if not track_obj:
            e = TrackNotFound(f"Could not retrieve track metadata for {link_track}")
            report_error(e, ids, link_track)
            raise e

        # If requested and provided via context, override with Spotify metadata for tagging
        if spotify_metadata and playlist_context and playlist_context.get('spotify_track_obj'):
            track_obj_for_tagging = playlist_context.get('spotify_track_obj')
        else:
            track_obj_for_tagging = track_obj

        # Set up download preferences
        preferences = Preferences()
        preferences.link = link_track
        preferences.song_metadata = track_obj_for_tagging  # Use selected track object (Spotify or Deezer) for tagging
        preferences.quality_download = quality_download
        preferences.output_dir = output_dir
        preferences.ids = ids
        preferences.recursive_quality = recursive_quality
        preferences.recursive_download = recursive_download
        preferences.not_interface = not_interface
        preferences.custom_dir_format = custom_dir_format
        preferences.custom_track_format = custom_track_format
        preferences.pad_tracks = pad_tracks
        preferences.initial_retry_delay = initial_retry_delay
        preferences.retry_delay_increase = retry_delay_increase
        preferences.max_retries = max_retries
        preferences.convert_to = convert_to
        preferences.bitrate = bitrate
        preferences.save_cover = save_cover
        preferences.market = market
        preferences.artist_separator = artist_separator
        preferences.spotify_metadata = bool(spotify_metadata)
        preferences.spotify_track_obj = playlist_context.get('spotify_track_obj') if (playlist_context and playlist_context.get('spotify_track_obj')) else None
        preferences.pad_number_width = pad_number_width

        if playlist_context:
            preferences.json_data = playlist_context.get('json_data')
            preferences.track_number = playlist_context.get('track_number')
            preferences.total_tracks = playlist_context.get('total_tracks')
            preferences.spotify_url = playlist_context.get('spotify_url')

        try:
            parent = 'playlist' if (playlist_context and playlist_context.get('json_data')) else None
            track = DW_TRACK(preferences, parent=parent).dw()
            return track
        except Exception as e:
            logger.error(f"Failed to download track: {str(e)}")
            report_error(e, ids, link_track)
            raise e

    def download_albumdee(
        self, link_album,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        make_zip=stock_zip,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        playlist_context=None,
        artist_separator: str = "; ",
        spotify_metadata: bool = False,
        spotify_album_obj=None,
        pad_number_width: int | str = 'auto'
    ) -> Album:

        link_is_valid(link_album)
        ids = get_ids(link_album)

        def report_error(e, current_ids, url):
            error_status = errorObject(ids=IDs(deezer=current_ids), error=str(e))
            callback_obj = albumCallbackObject(
                album=albumCbObject(title=f"Album ID {current_ids}", ids=IDs(deezer=current_ids)),
                status_info=error_status
            )
            report_progress(reporter=self.progress_reporter, callback_obj=callback_obj)

        try:
            # Get standardized album object
            album_obj = API.get_album(ids)
            if not album_obj:
                e = AlbumNotFound(f"Could not retrieve album metadata for {link_album}")
                report_error(e, ids, link_album)
                raise e
        except NoDataApi as e:
            report_error(e, ids, link_album)
            raise AlbumNotFound(link_album) from e

        # Set up download preferences
        preferences = Preferences()
        preferences.link = link_album
        preferences.song_metadata = album_obj  # Using the standardized album object
        preferences.quality_download = quality_download
        preferences.output_dir = output_dir
        preferences.ids = ids
        preferences.json_data = album_obj  # Pass the complete album object
        preferences.recursive_quality = recursive_quality
        preferences.recursive_download = recursive_download
        preferences.not_interface = not_interface
        preferences.make_zip = make_zip
        preferences.custom_dir_format = custom_dir_format
        preferences.custom_track_format = custom_track_format
        preferences.pad_tracks = pad_tracks
        preferences.initial_retry_delay = initial_retry_delay
        preferences.retry_delay_increase = retry_delay_increase
        preferences.max_retries = max_retries
        preferences.convert_to = convert_to
        preferences.bitrate = bitrate
        preferences.save_cover = save_cover
        preferences.market = market
        preferences.artist_separator = artist_separator
        preferences.spotify_metadata = bool(spotify_metadata)
        preferences.spotify_album_obj = spotify_album_obj
        preferences.pad_number_width = pad_number_width

        if playlist_context:
            preferences.json_data = playlist_context['json_data']
            preferences.track_number = playlist_context['track_number']
            preferences.total_tracks = playlist_context['total_tracks']
            preferences.spotify_url = playlist_context['spotify_url']

        try:
            album = DW_ALBUM(preferences).dw()
            return album
        except Exception as e:
            logger.error(f"Failed to download album: {str(e)}")
            report_error(e, ids, link_album)
            raise e

    def download_playlistdee(
        self, link_playlist,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        make_zip=stock_zip,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        artist_separator: str = "; ",
        pad_number_width: int | str = 'auto'
    ) -> Playlist:

        link_is_valid(link_playlist)
        ids = get_ids(link_playlist)

        playlist_obj = API.get_playlist(ids)
        if not playlist_obj:
            raise NoDataApi(f"Playlist {ids} not found.")

        # This part of fetching metadata track by track is now handled in __download__.py
        # The logic here is simplified to pass the full playlist object.

        preferences = Preferences()
        preferences.link = link_playlist
        # preferences.song_metadata is not needed here, DW_PLAYLIST will use json_data
        preferences.quality_download = quality_download
        preferences.output_dir = output_dir
        preferences.ids = ids
        preferences.json_data = playlist_obj
        preferences.recursive_quality = recursive_quality
        preferences.recursive_download = recursive_download
        preferences.not_interface = not_interface
        preferences.make_zip = make_zip
        preferences.custom_dir_format = custom_dir_format
        preferences.custom_track_format = custom_track_format
        preferences.pad_tracks = pad_tracks
        preferences.initial_retry_delay = initial_retry_delay
        preferences.retry_delay_increase = retry_delay_increase
        preferences.max_retries = max_retries
        preferences.convert_to = convert_to
        preferences.bitrate = bitrate
        preferences.save_cover = save_cover
        preferences.market = market
        preferences.artist_separator = artist_separator
        preferences.pad_number_width = pad_number_width

        playlist = DW_PLAYLIST(preferences).dw()

        return playlist

    def download_artisttopdee(
        self, link_artist,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        pad_number_width: int | str = 'auto'
    ) -> list[Track]:

        link_is_valid(link_artist)
        ids = get_ids(link_artist)

        # Assuming get_artist_top_tracks returns a list of track-like dicts with a 'link'
        top_tracks_json = API.get_artist_top_tracks(ids)['data']

        names = [
            self.download_trackdee(
                track['link'], output_dir,
                quality_download, recursive_quality,
                recursive_download, not_interface,
                custom_dir_format=custom_dir_format,
                custom_track_format=custom_track_format,
                pad_tracks=pad_tracks,
                convert_to=convert_to,
                bitrate=bitrate,
                save_cover=save_cover,
                market=market,
                pad_number_width=pad_number_width
            )
            for track in top_tracks_json
        ]

        return names

    def convert_spoty_to_dee_link_track(self, link_track):
        # Ensure Spotify session only when using spo->dee conversion
        self._ensure_spotify_session()
        link_is_valid(link_track)
        ids = get_ids(link_track)
        
        # Attempt via ISRC first
        track_json = Spo.get_track(ids)
        if not track_json:
            raise TrackNotFound(url=link_track, message="Spotify track metadata fetch failed.")
        external_ids = track_json.get('external_ids') or {}
        spo_isrc = (external_ids.get('isrc') or '').upper()
        spo_title = track_json.get('name', '')
        spo_album_title = (track_json.get('album') or {}).get('name', '')
        spo_tracknum = int(track_json.get('track_number') or 0)
        spo_artists = track_json.get('artists') or []
        spo_main_artist = (spo_artists[0].get('name') if spo_artists else '') or ''

        try:
            dz = API.get_track_json(f"isrc:{spo_isrc}")
            if dz and dz.get('id'):
                dz_json = dz
                tn = (dz_json.get('track_position') or dz_json.get('track_number') or 0)
                title_match = max(
                    _sim(_remove_parentheses(spo_title), _remove_parentheses(dz_json.get('title', ''))),
                    _sim(_remove_parentheses(spo_title), _remove_parentheses(dz_json.get('title_short', '')))
                )
                album_match = _sim(spo_album_title, (dz_json.get('album') or {}).get('title', ''))
                t_isrc = (dz_json.get('isrc') or '').upper()
                # Enforce ISRC match strictly in ISRC lookup path
                if (
                    t_isrc and spo_isrc and t_isrc == spo_isrc and
                    title_match >= 0.90 and album_match >= 0.90 and tn == spo_tracknum
                ):
                    return f"https://www.deezer.com/track/{dz_json.get('id')}"
        except Exception:
            pass
        
        # Fallback: search by title + artist + album
        query = f'"track:\'{spo_title}\' artist:\'{spo_main_artist}\' album:\'{spo_album_title}\'"'
        try:
            candidates = API.search_tracks_raw(query, limit=5)
        except Exception:
            candidates = []
        
        for cand in candidates:
            title_match = max(
                _sim(_remove_parentheses(spo_title), _remove_parentheses(cand.get('title', ''))),
                _sim(_remove_parentheses(spo_title), _remove_parentheses(cand.get('title_short', '')))
            )
            if title_match < 0.90:
                continue
            c_id = cand.get('id')
            if not c_id:
                continue
            try:
                dzc = API.get_track_json(str(c_id))
            except Exception:
                continue
            # Validate using track number and ISRC to be safe
            tn = (dzc.get('track_position') or dzc.get('track_number') or 0)
            if tn != spo_tracknum:
                continue
            t_isrc = (dzc.get('isrc') or '').upper()
            # Enforce ISRC strictly in fallback path as well: require present and equal
            if not spo_isrc or not t_isrc or t_isrc != spo_isrc:
                continue
            return f"https://www.deezer.com/track/{c_id}"
        
        raise TrackNotFound(url=link_track, message=f"Failed to find Deezer equivalent for ISRC {spo_isrc} from Spotify track {link_track}")

    def convert_isrc_to_dee_link_track(self, isrc_code: str) -> str:
        if not isinstance(isrc_code, str) or not isrc_code:
            raise ValueError("ISRC code must be a non-empty string.")

        isrc_query = f"isrc:{isrc_code}"
        logger.debug(f"Attempting Deezer track search with ISRC query: {isrc_query}")

        try:
            track_obj = API.get_track(isrc_query)
        except NoDataApi:
            msg = f"⚠ The track with ISRC '{isrc_code}' can't be found on Deezer :( ⚠"
            logger.warning(msg)
            raise TrackNotFound(url=f"isrc:{isrc_code}", message=msg)
        
        if not track_obj or not track_obj.type or not track_obj.ids or not track_obj.ids.deezer:
            msg = f"⚠ Deezer API returned no link for ISRC '{isrc_code}' :( ⚠"
            logger.warning(msg)
            raise TrackNotFound(url=f"isrc:{isrc_code}", message=msg)

        track_link_dee = f"https://www.deezer.com/{track_obj.type}/{track_obj.ids.deezer}"
        logger.info(f"Successfully converted ISRC {isrc_code} to Deezer link: {track_link_dee}")
        return track_link_dee

    def convert_spoty_to_dee_link_album(self, link_album):
        # Ensure Spotify session only when using spo->dee conversion
        self._ensure_spotify_session()
        link_is_valid(link_album)
        ids = get_ids(link_album)
        
        spotify_album_data = Spo.get_album(ids)
        if not spotify_album_data:
            raise AlbumNotFound(f"Failed to fetch Spotify album metadata for {link_album}")
        
        spo_album_title = spotify_album_data.get('name', '')
        spo_artists = spotify_album_data.get('artists') or []
        spo_main_artist = (spo_artists[0].get('name') if spo_artists else '') or ''
        external_ids = spotify_album_data.get('external_ids') or {}
        spo_upc = str(external_ids.get('upc') or '').strip().lstrip('0')
        
        # Try UPC first
        if spo_upc:
            try:
                dz_album = API.get_album_json(f"upc:{spo_upc}")
                if dz_album.get('id') and _sim(spo_album_title, dz_album.get('title', '')) >= 0.90:
                    return f"https://www.deezer.com/album/{dz_album.get('id')}"
            except Exception:
                pass
        
        # Fallback: title search
        q = f'"{spo_album_title}" {spo_main_artist}'.strip()
        try:
            candidates = API.search_albums_raw(q, limit=5)
        except Exception:
            candidates = []

        for cand in candidates:
            if _sim(spo_album_title, cand.get('title', '')) < 0.90:
                continue
            c_id = cand.get('id')
            if not c_id:
                continue
            try:
                dzc = API.get_album_json(str(c_id))
            except Exception:
                continue
            upc = str(dzc.get('upc') or '').strip().lstrip('0')
            if spo_upc and upc and spo_upc != upc:
                continue
            link_dee = f"https://www.deezer.com/album/{c_id}"
            return link_dee

        raise AlbumNotFound(f"Failed to convert Spotify album link {link_album} to a Deezer link after all attempts.")

    def download_trackspo(
        self, link_track,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        playlist_context=None,
        artist_separator: str = "; ",
        spotify_metadata: bool = False,
        pad_number_width: int | str = 'auto'
    ) -> Track:

        link_dee = self.convert_spoty_to_dee_link_track(link_track)

        # If requested, prepare Spotify track object for tagging in preferences via playlist_context
        if spotify_metadata:
            try:
                from deezspot.spotloader.__spo_api__ import tracking as spo_tracking
                spo_ids = get_ids(link_track)
                spo_track_obj = spo_tracking(spo_ids)
                if spo_track_obj:
                    if playlist_context is None:
                        playlist_context = {}
                    playlist_context = dict(playlist_context)
                    playlist_context['spotify_track_obj'] = spo_track_obj
            except Exception:
                pass

        track = self.download_trackdee(
            link_dee,
            output_dir=output_dir,
            quality_download=quality_download,
            recursive_quality=recursive_quality,
            recursive_download=recursive_download,
            not_interface=not_interface,
            custom_dir_format=custom_dir_format,
            custom_track_format=custom_track_format,
            pad_tracks=pad_tracks,
            initial_retry_delay=initial_retry_delay,
            retry_delay_increase=retry_delay_increase,
            max_retries=max_retries,
            convert_to=convert_to,
            bitrate=bitrate,
            save_cover=save_cover,
            market=market,
            playlist_context=playlist_context,
            artist_separator=artist_separator,
            spotify_metadata=spotify_metadata,
            pad_number_width=pad_number_width
        )

        return track

    def download_albumspo(
        self, link_album,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        make_zip=stock_zip,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        playlist_context=None,
        artist_separator: str = "; ",
        spotify_metadata: bool = False,
        spotify_album_obj=None,
        pad_number_width: int | str = 'auto'
    ) -> Album:

        link_dee = self.convert_spoty_to_dee_link_album(link_album)

        spotify_album_obj = None
        if spotify_metadata:
            # Only initialize Spotify session when we actually need Spotify metadata
            self._ensure_spotify_session()
            try:
                # Fetch full Spotify album with tracks once and convert to albumObject
                from deezspot.spotloader.__spo_api__ import tracking_album as spo_tracking_album
                spo_ids = get_ids(link_album)
                spotify_album_json = Spo.get_album(spo_ids)
                if spotify_album_json:
                    spotify_album_obj = spo_tracking_album(spotify_album_json)
            except Exception:
                spotify_album_obj = None

        album = self.download_albumdee(
            link_dee, output_dir,
            quality_download, recursive_quality,
            recursive_download, not_interface,
            make_zip, 
            custom_dir_format=custom_dir_format,
            custom_track_format=custom_track_format,
            pad_tracks=pad_tracks,
            initial_retry_delay=initial_retry_delay,
            retry_delay_increase=retry_delay_increase,
            max_retries=max_retries,
            convert_to=convert_to,
            bitrate=bitrate,
            save_cover=save_cover,
            market=market,
            playlist_context=playlist_context,
            artist_separator=artist_separator,
            spotify_metadata=spotify_metadata,
            spotify_album_obj=spotify_album_obj,
            pad_number_width=pad_number_width
        )

        return album

    def download_playlistspo(
        self, link_playlist,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        make_zip=stock_zip,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        artist_separator: str = "; ",
        spotify_metadata: bool = False,
        pad_number_width: int | str = 'auto'
    ) -> Playlist:

        link_is_valid(link_playlist)
        ids = get_ids(link_playlist)

        # Ensure Spotify session for fetching playlist and tracks
        self._ensure_spotify_session()

        playlist_json = Spo.get_playlist(ids)
        # Ensure we keep the playlist ID for callbacks
        if 'id' not in playlist_json:
            playlist_json['id'] = ids
        
        # Enrich items with full track objects so downstream expects Web API shape
        try:
            items = playlist_json.get('tracks', {}).get('items', []) or []
            track_ids = [it.get('track', {}).get('id') for it in items if it.get('track') and it['track'].get('id')]
            full = Spo.get_tracks(track_ids) if track_ids else {'tracks': []}
            full_list = full.get('tracks') or []
            full_by_id = {t.get('id'): t for t in full_list if t and t.get('id')}
            new_items = []
            for it in items:
                tid = (it.get('track') or {}).get('id')
                full_track = full_by_id.get(tid)
                if full_track:
                    new_items.append({'track': full_track})
                else:
                    new_items.append(it)
            playlist_json['tracks']['items'] = new_items
        except Exception:
            # If enrichment fails, continue with minimal ids
            pass

        
        # Extract track metadata for playlist callback object
        playlist_tracks_for_callback = []
        for item in playlist_json['tracks']['items']:
            if not item.get('track'):
                continue
            
            track_info = item['track']
            
            # Import the correct playlist-specific objects
            from deezspot.models.callback.playlist import (
                artistTrackPlaylistObject, 
                albumTrackPlaylistObject,
                artistAlbumTrackPlaylistObject,
                trackPlaylistObject
            )
            
            # Create artists with proper type
            track_artists = [artistTrackPlaylistObject(
                name=artist['name'],
                ids=IDs(spotify=artist.get('id'))
            ) for artist in track_info.get('artists', [])]
            
            # Process album with proper type and include images
            album_info = track_info.get('album', {})
            album_images = []
            if album_info.get('images'):
                album_images = [
                    {"url": img.get('url'), "height": img.get('height'), "width": img.get('width')}
                    for img in album_info.get('images', [])
                ]
            
            # Process album artists
            album_artists = []
            if album_info.get('artists'):
                album_artists = [
                    artistAlbumTrackPlaylistObject(
                        name=artist.get('name'),
                        ids=IDs(spotify=artist.get('id'))
                    )
                    for artist in album_info.get('artists', [])
                ]
            
            album_obj = albumTrackPlaylistObject(
                title=album_info.get('name', 'Unknown Album'),
                ids=IDs(spotify=album_info.get('id')),
                images=album_images,
                artists=album_artists,
                album_type=album_info.get('album_type', ''),
                release_date={
                    "year": int(album_info.get('release_date', '0').split('-')[0]) if album_info.get('release_date') else 0,
                    "month": int(album_info.get('release_date', '0-0').split('-')[1]) if album_info.get('release_date') and len(album_info.get('release_date').split('-')) > 1 else 0,
                    "day": int(album_info.get('release_date', '0-0-0').split('-')[2]) if album_info.get('release_date') and len(album_info.get('release_date').split('-')) > 2 else 0
                },
                total_tracks=album_info.get('total_tracks', 0)
            )
            
            # Create track with proper playlist-specific type
            track_obj = trackPlaylistObject(
                title=track_info.get('name', 'Unknown Track'),
                artists=track_artists,
                album=album_obj,
                duration_ms=track_info.get('duration_ms', 0),
                explicit=track_info.get('explicit', False),
                ids=IDs(
                    spotify=track_info.get('id'), 
                    isrc=track_info.get('external_ids', {}).get('isrc')
                ),
                disc_number=track_info.get('disc_number', 1),
                track_number=track_info.get('track_number', 0)
            )
            playlist_tracks_for_callback.append(track_obj)
        
        playlist_obj = playlistCbObject(
            title=playlist_json['name'],
            owner=userObject(name=playlist_json.get('owner', {}).get('display_name', 'Unknown Owner')),
            ids=IDs(spotify=playlist_json.get('id', ids)),
            tracks=playlist_tracks_for_callback  # Populate tracks array with track objects
        )

        status_obj_init = initializingObject(ids=playlist_obj.ids)
        callback_obj_init = playlistCallbackObject(playlist=playlist_obj, status_info=status_obj_init)
        report_progress(reporter=self.progress_reporter, callback_obj=callback_obj_init)

        total_tracks = playlist_json['tracks']['total']
        playlist_tracks = playlist_json['tracks']['items']
        playlist = Playlist()
        tracks = playlist.tracks

        successful_tracks_cb = []
        failed_tracks_cb = []
        skipped_tracks_cb = []

        for index, item in enumerate(playlist_tracks, 1):
            is_track = item.get('track')
            if not is_track:
                logger.warning(f"Skipping an item in playlist {playlist_obj.title} as it's not a valid track (likely unavailable in region).")
                unknown_track = trackCbObject(title="Unknown Skipped Item", artists=[artistTrackObject(name="")])
                reason = "Playlist item was not a valid track object or is not available in your region."
                
                failed_tracks_cb.append(failedTrackObject(track=unknown_track, reason=reason))
                
                # Create a placeholder for the failed item
                failed_track = Track(
                    tags={'music': 'Unknown Skipped Item', 'artist': 'Unknown'},
                    song_path=None, file_format=None, quality=None, link=None, ids=None
                )
                failed_track.success = False
                failed_track.error_message = reason
                tracks.append(failed_track)
                continue

            track_info = is_track
            track_name = track_info.get('name', 'Unknown Track')
            artist_name = track_info['artists'][0]['name'] if track_info.get('artists') else 'Unknown Artist'
            link_track = track_info.get('external_urls', {}).get('spotify')
            if not link_track:
                tid = track_info.get('id')
                if tid:
                    link_track = f"https://open.spotify.com/track/{tid}"

            if not link_track:
                logger.warning(f"The track \"{track_name}\" is not available on Spotify :(")
                continue

            try:
                playlist_ctx = {
                    'json_data': playlist_json,
                    'track_number': index,
                    'total_tracks': total_tracks,
                    'spotify_url': link_track
                }
                # Attach Spotify track object for tagging if requested
                if False: # placeholder, will be handled via spotify_metadata in download_trackspo
                    pass
                downloaded_track = self.download_trackspo(
                    link_track,
                    output_dir=output_dir, quality_download=quality_download,
                    recursive_quality=recursive_quality, recursive_download=recursive_download,
                    not_interface=not_interface, custom_dir_format=custom_dir_format,
                    custom_track_format=custom_track_format, pad_tracks=pad_tracks,
                    initial_retry_delay=initial_retry_delay, retry_delay_increase=retry_delay_increase,
                    max_retries=max_retries, convert_to=convert_to, bitrate=bitrate,
                    save_cover=save_cover, market=market, playlist_context=playlist_ctx,
                    artist_separator=artist_separator, spotify_metadata=False,
                    pad_number_width=pad_number_width
                )
                tracks.append(downloaded_track)
                
                # After download, check status for summary
                if getattr(downloaded_track, 'was_skipped', False):
                    skipped_tracks_cb.append(playlist_obj.tracks[index-1])
                elif downloaded_track.success:
                    successful_tracks_cb.append(playlist_obj.tracks[index-1])
                else:
                    failed_tracks_cb.append(failedTrackObject(track=playlist_obj.tracks[index-1], reason=getattr(downloaded_track, 'error_message', 'Unknown reason')))
            except Exception as e:
                logger.error(f"Track '{track_name}' in playlist '{playlist_obj.title}' failed: {e}")
                failed_tracks_cb.append(failedTrackObject(track=playlist_obj.tracks[index-1], reason=str(e)))
                current_track_object = Track({'music': track_name, 'artist': artist_name}, None, None, None, link_track, None)
                current_track_object.success = False
                current_track_object.error_message = str(e)
                tracks.append(current_track_object)

        # Finalize summary and callbacks (existing logic continues below in file)...

        total_from_spotify = playlist_json['tracks']['total']
        processed_count = len(successful_tracks_cb) + len(skipped_tracks_cb) + len(failed_tracks_cb)

        if total_from_spotify != processed_count:
            logger.warning(
                f"Playlist '{playlist_obj.title}' metadata reports {total_from_spotify} tracks, "
                f"but only {processed_count} were processed. This might indicate that not all pages of tracks were retrieved from Spotify."
            )

        from deezspot.libutils.write_m3u import write_tracks_to_m3u
        m3u_path = write_tracks_to_m3u(output_dir, playlist_obj.title, tracks)

        summary_obj = summaryObject(
            successful_tracks=successful_tracks_cb,
            skipped_tracks=skipped_tracks_cb,
            failed_tracks=failed_tracks_cb,
            total_successful=len(successful_tracks_cb),
            total_skipped=len(skipped_tracks_cb),
            total_failed=len(failed_tracks_cb)
        )
        # Include m3u path in summary and callback
        summary_obj.m3u_path = m3u_path
        status_obj_done = doneObject(ids=playlist_obj.ids, summary=summary_obj)
        callback_obj_done = playlistCallbackObject(playlist=playlist_obj, status_info=status_obj_done)
        report_progress(reporter=self.progress_reporter, callback_obj=callback_obj_done)

        if make_zip:
            zip_name = f"{output_dir}/playlist_{sanitize_name(playlist_obj.title)}.zip"
            create_zip(tracks, zip_name=zip_name)
            playlist.zip_path = zip_name

        return playlist

    def download_name(
        self, artist, song,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        custom_dir_format=None,
        custom_track_format=None,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        pad_tracks=True,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        artist_separator: str = "; ",
        pad_number_width: int | str = 'auto'
    ) -> Track:

        query = f"track:{song} artist:{artist}"
        search = self.__spo.search(query)
        
        items = search['tracks']['items']

        if len(items) == 0:
            msg = f"No result for {query} :("
            raise TrackNotFound(message=msg)

        link_track = items[0]['external_urls']['spotify']

        track = self.download_trackspo(
            link_track,
            output_dir=output_dir,
            quality_download=quality_download,
            recursive_quality=recursive_quality,
            recursive_download=recursive_download,
            not_interface=not_interface,
            custom_dir_format=custom_dir_format,
            custom_track_format=custom_track_format,
            pad_tracks=pad_tracks,
            initial_retry_delay=initial_retry_delay,
            retry_delay_increase=retry_delay_increase,
            max_retries=max_retries,
            convert_to=convert_to,
            bitrate=bitrate,
            save_cover=save_cover,
            market=market,
            artist_separator=artist_separator,
            pad_number_width=pad_number_width
        )

        return track

    def download_episode(
        self,
        link_episode,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        artist_separator: str = "; "
    ) -> Episode:
        
        logger.warning("Episode download logic is not fully refactored and might not work as expected with new reporting.")
        link_is_valid(link_episode)
        ids = get_ids(link_episode)
        
        try:
            # This will likely fail as API.tracking is gone.
            episode_metadata = API.get_episode(ids)
        except (NoDataApi, MarketAvailabilityError) as e:
            raise TrackNotFound(url=link_episode, message=f"Episode not available: {e}") from e
        except Exception:
            # Fallback to GW API if public API fails for any reason
            infos = self.__gw_api.get_episode_data(ids)
            if not infos:
                raise TrackNotFound(f"Episode {ids} not found")
            episode_metadata = {
                'music': infos.get('EPISODE_TITLE', ''), 'artist': infos.get('SHOW_NAME', ''),
                'album': infos.get('SHOW_NAME', ''), 'date': infos.get('EPISODE_PUBLISHED_TIMESTAMP', '').split()[0],
                'genre': 'Podcast', 'explicit': infos.get('SHOW_IS_EXPLICIT', '2'),
                'disc': 1, 'track': 1, 'duration': int(infos.get('DURATION', 0)), 'isrc': None,
                'image': infos.get('EPISODE_IMAGE_MD5', '')
            }

        preferences = Preferences()
        preferences.link = link_episode
        preferences.song_metadata = episode_metadata
        preferences.quality_download = quality_download
        preferences.output_dir = output_dir
        preferences.ids = ids
        preferences.recursive_quality = recursive_quality
        preferences.recursive_download = recursive_download
        preferences.not_interface = not_interface
        preferences.max_retries = max_retries
        preferences.convert_to = convert_to
        preferences.bitrate = bitrate
        preferences.save_cover = save_cover
        preferences.is_episode = True
        preferences.market = market
        preferences.artist_separator = artist_separator

        episode = DW_EPISODE(preferences).dw()

        return episode
        
    def download_smart(
        self, link,
        output_dir=stock_output,
        quality_download=stock_quality,
        recursive_quality=stock_recursive_quality,
        recursive_download=stock_recursive_download,
        not_interface=stock_not_interface,
        make_zip=stock_zip,
        custom_dir_format=None,
        custom_track_format=None,
        pad_tracks=True,
        initial_retry_delay=30,
        retry_delay_increase=30,
        max_retries=5,
        convert_to=None,
        bitrate=None,
        save_cover=stock_save_cover,
        market=stock_market,
        artist_separator: str = "; "
    ) -> Smart:

        link_is_valid(link)
        link = what_kind(link)
        smart = Smart()

        if "spotify.com" in link:
            source = "spotify"
        elif "deezer.com" in link:
            source = "deezer"
        else:
            raise InvalidLink(link)

        smart.source = source
        
        # Smart download reporting can be enhanced later if needed
        # For now, the individual download functions will do the reporting.

        if "track/" in link:
            func = self.download_trackspo if source == 'spotify' else self.download_trackdee
            track = func(
                link, output_dir=output_dir, quality_download=quality_download,
                recursive_quality=recursive_quality, recursive_download=recursive_download,
                not_interface=not_interface, custom_dir_format=custom_dir_format,
                custom_track_format=custom_track_format, pad_tracks=pad_tracks,
                initial_retry_delay=initial_retry_delay, retry_delay_increase=retry_delay_increase,
                max_retries=max_retries, convert_to=convert_to, bitrate=bitrate,
                save_cover=save_cover, market=market, artist_separator=artist_separator
            )
            smart.type = "track"
            smart.track = track

        elif "album/" in link:
            func = self.download_albumspo if source == 'spotify' else self.download_albumdee
            album = func(
                link, output_dir=output_dir, quality_download=quality_download,
                recursive_quality=recursive_quality, recursive_download=recursive_download,
                not_interface=not_interface, make_zip=make_zip,
                custom_dir_format=custom_dir_format, custom_track_format=custom_track_format,
                pad_tracks=pad_tracks, initial_retry_delay=initial_retry_delay,
                retry_delay_increase=retry_delay_increase, max_retries=max_retries,
                convert_to=convert_to, bitrate=bitrate, save_cover=save_cover,
                market=market, artist_separator=artist_separator
            )
            smart.type = "album"
            smart.album = album

        elif "playlist/" in link:
            func = self.download_playlistspo if source == 'spotify' else self.download_playlistdee
            playlist = func(
                link, output_dir=output_dir, quality_download=quality_download,
                recursive_quality=recursive_quality, recursive_download=recursive_download,
                not_interface=not_interface, make_zip=make_zip,
                custom_dir_format=custom_dir_format, custom_track_format=custom_track_format,
                pad_tracks=pad_tracks, initial_retry_delay=initial_retry_delay,
                retry_delay_increase=retry_delay_increase, max_retries=max_retries,
                convert_to=convert_to, bitrate=bitrate, save_cover=save_cover,
                market=market, artist_separator=artist_separator
            )
            smart.type = "playlist"
            smart.playlist = playlist

        return smart



#!/usr/bin/python3

from typing import Optional, Any, Dict, List
from deezspot.libutils.logging_utils import report_progress
from deezspot.models.callback import (
    trackCallbackObject, albumCallbackObject, playlistCallbackObject,
    initializingObject, skippedObject, retryingObject, realTimeObject, 
    errorObject, doneObject, summaryObject
)


def _get_reporter():
    """Get the active progress reporter from Download_JOB"""
    # Import here to avoid circular imports
    try:
        from deezspot.spotloader.__download__ import Download_JOB as SpotifyDJ
        if hasattr(SpotifyDJ, 'progress_reporter') and SpotifyDJ.progress_reporter:
            return SpotifyDJ.progress_reporter
    except ImportError:
        pass
    
    try:
        from deezspot.deezloader.__download__ import Download_JOB as DeezerDJ
        if hasattr(DeezerDJ, 'progress_reporter') and DeezerDJ.progress_reporter:
            return DeezerDJ.progress_reporter
    except ImportError:
        pass
    
    return None


def report_track_initializing(
    track_obj: Any,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None
) -> None:
    """
    Report track initialization status.
    
    Args:
        track_obj: Track object being initialized
        preferences: Preferences object with convert_to/bitrate info
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
    """
    status_obj = initializingObject(
        ids=getattr(track_obj, 'ids', None),
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None)
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_track_skipped(
    track_obj: Any,
    reason: str,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None
) -> None:
    """
    Report track skipped status.
    
    Args:
        track_obj: Track object being skipped
        reason: Reason for skipping
        preferences: Preferences object with convert_to/bitrate info
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
    """
    status_obj = skippedObject(
        ids=getattr(track_obj, 'ids', None),
        reason=reason,
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None)
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_track_retrying(
    track_obj: Any,
    retry_count: int,
    seconds_left: int,
    error: str,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None
) -> None:
    """
    Report track retry status.
    
    Args:
        track_obj: Track object being retried
        retry_count: Current retry attempt number
        seconds_left: Seconds until next retry
        error: Error that caused the retry
        preferences: Preferences object with convert_to/bitrate info
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
    """
    status_obj = retryingObject(
        ids=getattr(track_obj, 'ids', None),
        retry_count=retry_count,
        seconds_left=seconds_left,
        error=error,
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None)
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_track_realtime_progress(
    track_obj: Any,
    time_elapsed: int,
    progress: int,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None
) -> None:
    """
    Report real-time track download progress.
    
    Args:
        track_obj: Track object being downloaded
        time_elapsed: Milliseconds elapsed
        progress: Progress percentage (0-100)
        preferences: Preferences object with convert_to/bitrate info
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
    """
    status_obj = realTimeObject(
        ids=getattr(track_obj, 'ids', None),
        time_elapsed=time_elapsed,
        progress=progress,
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None)
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_track_error(
    track_obj: Any,
    error: str,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None
) -> None:
    """
    Report track error status.
    
    Args:
        track_obj: Track object that errored
        error: Error message
        preferences: Preferences object with convert_to/bitrate info
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
    """
    status_obj = errorObject(
        ids=getattr(track_obj, 'ids', None),
        error=error,
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None)
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_track_done(
    track_obj: Any,
    preferences: Any,
    summary: Optional[summaryObject] = None,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None,
    *,
    final_path: Optional[str] = None,
    download_quality: Optional[str] = None
) -> None:
    """
    Report track completion status.
    
    Args:
        track_obj: Track object that completed
        preferences: Preferences object with convert_to/bitrate info
        summary: Optional summary object for single track downloads
        parent_obj: Parent object (album/playlist) if applicable
        current_track: Current track number for progress
        total_tracks: Total tracks for progress
        final_path: Final filesystem path of the produced file
        download_quality: String label of the used download quality (e.g., OGG_160, OGG_320 or FLAC/MP3_320)
    """
    status_obj = doneObject(
        ids=getattr(track_obj, 'ids', None),
        summary=summary,
        convert_to=getattr(preferences, 'convert_to', None),
        bitrate=getattr(preferences, 'bitrate', None),
        final_path=final_path,
        download_quality=download_quality
    )
    
    callback_obj = trackCallbackObject(
        track=track_obj,
        status_info=status_obj,
        current_track=current_track or getattr(preferences, 'track_number', None),
        total_tracks=total_tracks,
        parent=parent_obj
    )
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_album_initializing(album_obj: Any) -> None:
    """
    Report album initialization status.
    
    Args:
        album_obj: Album object being initialized
    """
    status_obj = initializingObject(ids=getattr(album_obj, 'ids', None))
    callback_obj = albumCallbackObject(album=album_obj, status_info=status_obj)
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_album_done(album_obj: Any, summary: summaryObject) -> None:
    """
    Report album completion status.
    
    Args:
        album_obj: Album object that completed
        summary: Summary of track download results
    """
    status_obj = doneObject(ids=getattr(album_obj, 'ids', None), summary=summary)
    callback_obj = albumCallbackObject(album=album_obj, status_info=status_obj)
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_playlist_initializing(playlist_obj: Any) -> None:
    """
    Report playlist initialization status.
    
    Args:
        playlist_obj: Playlist object being initialized
    """
    status_obj = initializingObject(ids=getattr(playlist_obj, 'ids', None))
    callback_obj = playlistCallbackObject(playlist=playlist_obj, status_info=status_obj)
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


def report_playlist_done(playlist_obj: Any, summary: summaryObject, *, m3u_path: Optional[str] = None) -> None:
    """
    Report playlist completion status.
    
    Args:
        playlist_obj: Playlist object that completed
        summary: Summary of track download results
        m3u_path: Final path of the generated m3u file, if any
    """
    if m3u_path:
        summary.m3u_path = m3u_path
    status_obj = doneObject(ids=getattr(playlist_obj, 'ids', None), summary=summary)
    callback_obj = playlistCallbackObject(playlist=playlist_obj, status_info=status_obj)
    
    reporter = _get_reporter()
    if reporter:
        report_progress(reporter=reporter, callback_obj=callback_obj)


# Convenience function for generic track reporting
def report_track_status(
    status_type: str,
    track_obj: Any,
    preferences: Any,
    parent_obj: Optional[Any] = None,
    current_track: Optional[int] = None,
    total_tracks: Optional[int] = None,
    **kwargs
) -> None:
    """
    Generic track status reporting function.
    
    Args:
        status_type: Type of status ('initializing', 'skipped', 'retrying', 'error', 'done', 'realtime')
        track_obj: Track object
        preferences: Preferences object
        parent_obj: Parent object if applicable
        current_track: Current track number
        total_tracks: Total tracks
        **kwargs: Additional parameters specific to status type
    """
    if status_type == 'initializing':
        report_track_initializing(track_obj, preferences, parent_obj, current_track, total_tracks)
    elif status_type == 'skipped':
        report_track_skipped(track_obj, kwargs.get('reason', 'Unknown'), preferences, parent_obj, current_track, total_tracks)
    elif status_type == 'retrying':
        report_track_retrying(track_obj, kwargs.get('retry_count', 0), kwargs.get('seconds_left', 0), 
                            kwargs.get('error', 'Unknown'), preferences, parent_obj, current_track, total_tracks)
    elif status_type == 'error':
        report_track_error(track_obj, kwargs.get('error', 'Unknown'), preferences, parent_obj, current_track, total_tracks)
    elif status_type == 'done':
        report_track_done(
            track_obj,
            preferences,
            kwargs.get('summary'),
            parent_obj,
            current_track,
            total_tracks,
            final_path=kwargs.get('final_path'),
            download_quality=kwargs.get('download_quality')
        )
    elif status_type == 'realtime':
        report_track_realtime_progress(track_obj, kwargs.get('time_elapsed', 0), kwargs.get('progress', 0),
                                     preferences, parent_obj, current_track, total_tracks) 


#!/usr/bin/python3

import re
from unicodedata import normalize
from os import makedirs
from datetime import datetime
from urllib.parse import urlparse
from requests import get as req_get
from zipfile import ZipFile, ZIP_DEFLATED
from deezspot.models.download.track import Track
from deezspot.exceptions import InvalidLink
from deezspot.libutils.others_settings import supported_link, header
from deezspot.libutils.logging_utils import ProgressReporter, logger

from os.path import (
    isdir, basename,
    join, isfile, dirname
)

def link_is_valid(link):
    netloc = urlparse(link).netloc

    if not any(
        c_link == netloc
        for c_link in supported_link
    ):
        raise InvalidLink(link)

def get_ids(link):
    parsed = urlparse(link)
    path = parsed.path
    ids = path.split("/")[-1]

    return ids

def request(url):
    thing = req_get(url, headers=header)
    return thing

def __check_dir(directory):
    if not isdir(directory):
        makedirs(directory)

def sanitize_name(string, max_length=200):
    """Sanitize a string for use as a filename or directory name.
    
    This version maps filesystem-conflicting ASCII characters to Unicode
    lookalikes (mostly fullwidth forms) rather than dropping or replacing
    with ASCII fallbacks. This preserves readability while avoiding
    path-separator or Windows-invalid characters.
    
    Args:
        string: The string to sanitize
        max_length: Maximum length for the resulting string
        
    Returns:
        A sanitized string safe for use in file paths
    """
    if string is None:
        return "Unknown"
        
    # Convert to string if not already
    string = str(string)

    # Map invalid/reserved characters to Unicode fullwidth or similar lookalikes
    # to avoid filesystem conflicts while keeping readability.
    # Windows-invalid: < > : " / \ | ? * and control chars
    replacements = {
        "\\": "＼",  # U+FF3C FULLWIDTH REVERSE SOLIDUS
        "/": "／",   # U+FF0F FULLWIDTH SOLIDUS
        ":": "：",   # U+FF1A FULLWIDTH COLON
        "*": "＊",   # U+FF0A FULLWIDTH ASTERISK
        "?": "？",   # U+FF1F FULLWIDTH QUESTION MARK
        "\"": "＂",  # U+FF02 FULLWIDTH QUOTATION MARK
        "<": "＜",   # U+FF1C FULLWIDTH LESS-THAN SIGN
        ">": "＞",   # U+FF1E FULLWIDTH GREATER-THAN SIGN
        "|": "｜",   # U+FF5C FULLWIDTH VERTICAL LINE
        "&": "＆",   # U+FF06 FULLWIDTH AMPERSAND
        "$": "＄",   # U+FF04 FULLWIDTH DOLLAR SIGN
        ";": "；",   # U+FF1B FULLWIDTH SEMICOLON
        "\t": " ",  # Tab to space
        "\n": " ",  # Newline to space
        "\r": " ",  # Carriage return to space
        "\0": "",   # Null byte removed
    }
    
    for old, new in replacements.items():
        string = string.replace(old, new)

    # Remove any other non-printable characters
    string = ''.join(char for char in string if char.isprintable())

    # Remove leading/trailing whitespace
    string = string.strip()

    # Replace multiple spaces with a single space
    string = re.sub(r'\s+', ' ', string)

    # Truncate if too long
    if len(string) > max_length:
        string = string[:max_length]
        
    # Ensure we don't end with a dot or space (can cause issues in some filesystems)
    string = string.rstrip('. ')

    # Provide a fallback for empty strings
    if not string:
        string = "Unknown"

    # Normalize to NFC to keep composed characters stable but avoid
    # compatibility decomposition that might revert fullwidth mappings.
    string = normalize('NFC', string)
        
    return string

# Keep the original function name for backward compatibility
def var_excape(string):
    """Legacy function name for backward compatibility."""
    return sanitize_name(string)

def convert_to_date(date: str):
    if date == "0000-00-00":
        date = "0001-01-01"
    elif date.isdigit():
        date = f"{date}-01-01"
    date = datetime.strptime(date, "%Y-%m-%d")
    return date

def what_kind(link):
    url = request(link).url
    if url.endswith("/"):
        url = url[:-1]
    return url

def __get_tronc(string):
    return string[:len(string) - 1]

def apply_custom_format(format_str, metadata: dict, pad_tracks=True, pad_number_width: int | None = None) -> str:
    def replacer(match):
        full_key = match.group(1)  # e.g., "artist", "ar_album_1"

        # Allow custom artist/album-artist separator to be provided via metadata
        separator = metadata.get('artist_separator', ';')
        if not isinstance(separator, str) or separator == "":
            separator = ';'

        # Check for specific indexed placeholders: artist_INDEX or ar_album_INDEX
        # Allows %artist_1%, %ar_album_1%, etc.
        indexed_artist_match = re.fullmatch(r'(artist|ar_album)_(\d+)', full_key)

        if indexed_artist_match:
            base_key = indexed_artist_match.group(1)  # "artist" or "ar_album"
            try:
                index = int(indexed_artist_match.group(2))
            except ValueError: # Should not happen with \d+ but good practice
                return ""


            raw_value = metadata.get(base_key)  # Get the value of "artist" or "ar_album"
            items = []

            if isinstance(raw_value, str):
                # Split by provided separator and strip whitespace
                items = [item.strip() for item in raw_value.split(separator) if item.strip()]
            elif isinstance(raw_value, list):
                # Convert all items to string, strip whitespace
                items = [str(item).strip() for item in raw_value if str(item).strip()]
            # If raw_value is not string or list, items remains []

            if items:  # If we have a list of artists/ar_album
                if 1 <= index <= len(items):
                    return items[index - 1]
                elif items:  # Index out of bounds, but list is not empty
                    return items[0]  # Fallback to the first item
                # If items is empty after processing, fall through
            
            # Fallback if no items or base_key was not found or not list/string
            return ""

        else:
            # Original non-indexed placeholder logic (for %album%, %title%, %artist%, %ar_album%, etc.)
            value = metadata.get(full_key, '')

            # Handle None values safely
            if value is None:
                if full_key in ['tracknum', 'discnum', 'playlistnum']:
                    value = '1' if full_key == 'discnum' else '0'
                else:
                    value = ''

            if full_key == 'year' and value:
                if isinstance(value, datetime):
                    return str(value.year)
                # Fallback for string-based dates like "YYYY-MM-DD" or just "YYYY"
                return str(value).split('-')[0]

            if pad_tracks and full_key in ['tracknum', 'discnum', 'playlistnum']:
                str_value = str(value)
                if str_value.isdigit():
                    if isinstance(pad_number_width, int) and pad_number_width >= 1:
                        return str_value.zfill(pad_number_width)
                    # Default legacy behavior: pad single digits to width 2
                    if len(str_value) == 1:
                        return str_value.zfill(2)
                return str_value
            return str(value)

    return re.sub(r'%([^%]+)%', replacer, format_str)

def __get_dir(song_metadata, output_dir, custom_dir_format=None, pad_tracks=True, pad_number_width: int | None = None):
    # If custom_dir_format is explicitly empty or None, use output_dir directly
    if not custom_dir_format:
        # Ensure output_dir itself exists, as __check_dir won't be called on a subpath
        __check_dir(output_dir)
        return output_dir

    # Apply formatting per path component so only slashes from the format
    # create directories; slashes from data are sanitized inside components.
    format_parts = custom_dir_format.split("/")
    formatted_parts = [
        apply_custom_format(part, song_metadata, pad_tracks, pad_number_width) for part in format_parts
    ]
    sanitized_path_segment = "/".join(
        sanitize_name(part) for part in formatted_parts
    )

    # Join with the base output directory
    path = join(output_dir, sanitized_path_segment)
    
    # __check_dir will create the directory if it doesn't exist.
    __check_dir(path)
    return path

def set_path(
    song_metadata, output_dir,
    song_quality, file_format,
    is_episode=False,
    custom_dir_format=None,
    custom_track_format=None,
    pad_tracks=True,
    pad_number_width: int | None = None
):
    # Determine the directory for the song
    directory = __get_dir(
        song_metadata,
        output_dir,
        custom_dir_format=custom_dir_format,
        pad_tracks=pad_tracks,
        pad_number_width=pad_number_width
    )

    # Determine the filename for the song
    # Default track format if no custom one is provided
    if custom_track_format is None:
        if is_episode:
            # Default for episodes: %music%
            custom_track_format = "%music%"
        else:
            # Default for tracks: %artist% - %music%
            custom_track_format = "%artist% - %music%"
    
    # Prepare metadata for formatting, including quality if available
    effective_metadata = dict(song_metadata) # Create a mutable copy
    if song_quality:
        effective_metadata['quality'] = f"[{song_quality}]"
    # else: if song_quality is None or empty, 'quality' won't be in effective_metadata,
    # so %quality% placeholder will be replaced by an empty string by apply_custom_format.

    # Apply the custom format string for the track filename.
    # pad_tracks is passed along for track/disc numbers in filename.
    track_filename_base = apply_custom_format(custom_track_format, effective_metadata, pad_tracks, pad_number_width)
    track_filename_base = sanitize_name(track_filename_base)

    # Add file format (extension) to the filename
    if file_format:
        ext = file_format if file_format.startswith('.') else f".{file_format}"
        filename = f"{track_filename_base}{ext}"
    else: # No file_format provided (should not happen for standard audio, but handle defensively)
        filename = track_filename_base

    return join(directory, filename)

def create_zip(
    tracks: list[Track],
    output_dir=None,
    song_metadata=None, # Album/Playlist level metadata
    song_quality=None, # Overall quality for the zip, if applicable
    zip_name=None, # Specific name for the zip file
    custom_dir_format=None # To determine zip name if not provided, and for paths inside zip
):
    # Determine the zip file name and path
    if zip_name:
        # If zip_name is a full path, use it as is.
        # Otherwise, prepend output_dir.
        if not basename(zip_name) == zip_name: # Checks if it's just a filename
            actual_zip_path = zip_name
        else:
            # Ensure output_dir exists for placing the zip file
            if not output_dir:
                # Fallback to a default if output_dir is not provided with a relative zip_name
                output_dir = "."
                __check_dir(output_dir)
            actual_zip_path = join(output_dir, zip_name)
    elif song_metadata and output_dir: # Construct default name if song_metadata and output_dir exist
        # Use album/playlist name and quality for default zip name
        # Sanitize the album/playlist name part of the zip file
        name_part = sanitize_name(song_metadata.get('album', song_metadata.get('name', 'archive')))
        quality_part = f" [{song_quality}]" if song_quality else ""
        actual_zip_path = join(output_dir, f"{name_part}{quality_part}.zip")
    else:
        # Fallback zip name if not enough info
        actual_zip_path = join(output_dir if output_dir else ".", "archive.zip")

    # Ensure the directory for the zip file exists
    zip_dir = dirname(actual_zip_path)
    __check_dir(zip_dir)

    with ZipFile(actual_zip_path, 'w', ZIP_DEFLATED) as zf:
        for track in tracks:
            if track.success and isfile(track.song_path):
                # Determine path inside the zip
                # This uses the same logic as saving individual files,
                # but relative to the zip root.
                # We pass an empty string as base_output_dir to set_path essentially,
                # so it generates a relative path structure.
                path_in_zip = set_path(
                    track.tags, # Use individual track metadata for path inside zip
                    "",         # Base output dir (empty for relative paths in zip)
                    track.quality,
                    track.file_format,
                    custom_dir_format=custom_dir_format, # Use album/playlist custom dir format
                    custom_track_format=track.tags.get('custom_track_format'), # Use track specific if available
                    pad_tracks=track.tags.get('pad_tracks', True)
                )
                # Remove leading slash if any, to ensure it's relative inside zip
                path_in_zip = path_in_zip.lstrip('/').lstrip('\\')
                
                zf.write(track.song_path, arcname=path_in_zip)
    return actual_zip_path

def trasform_sync_lyric(lyric):
    sync_array = []
    for a in lyric:
        if "milliseconds" in a:
            arr = (a['line'], int(a['milliseconds']))
            sync_array.append(arr)
    return sync_array

def save_cover_image(image_data: bytes, directory_path: str, cover_filename: str = "cover.jpg"):
    if not image_data:
        logger.warning(f"No image data provided to save cover in {directory_path}.")
        return

    if not isdir(directory_path):
        # This case should ideally be handled by prior directory creation (e.g., __get_dir)
        # but as a fallback, we can try to create it or log a warning.
        logger.warning(f"Directory {directory_path} does not exist. Attempting to create it for cover image.")
        try:
            makedirs(directory_path, exist_ok=True)
            logger.info(f"Created directory {directory_path} for cover image.")
        except OSError as e:
            logger.error(f"Failed to create directory {directory_path} for cover: {e}")
            return

    cover_path = join(directory_path, cover_filename)
    try:
        with open(cover_path, "wb") as f:
            f.write(image_data)
        logger.info(f"Successfully saved cover image to {cover_path}")
    except OSError as e:
        logger.error(f"Failed to save cover image to {cover_path}: {e}")
